/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * fonts.c
 *
 *  Created on: 22/10/2010
 *      Author: raster
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>

#include "fonts.h"
#include "minibus/minibus.h"

struct _font_base *fonts_init(int8_t *font_path) {

	struct _font_base *object;
#ifdef USE_ARM
	struct stat font_data;
#else
	struct stat64 font_data;
#endif

	uint32_t *pointer;

	if (-1==stat64((char *)font_path,&font_data)) {
		return NULL;
	}

	object=malloc(sizeof(struct _font_base));
	if (!object) {
		return NULL;
	}
	object->fd=open((char *)font_path,O_RDONLY);

	if (object->fd<0) {
		free(object);
		return NULL;
	}

	object->size=font_data.st_size;
	object->data=mmap(NULL,font_data.st_size,PROT_READ,MAP_SHARED,object->fd,0);
	if (!object->data) {
		free(object);
		return NULL;
	}
	object->path=(int8_t *)strdup((char *)font_path);
	pointer=(uint32_t *)object->data;
	object->average_height=*pointer;
	object->first_glyph=*(pointer+1);
	object->last_glyph=*(pointer+2);
	object->offsets=(uint32_t *)(object->data+12);
	object->glyphs=object->data+16+4*(object->last_glyph-object->first_glyph);
	object->r=0;
	object->g=0;
	object->b=0;
	memset(object->colors,0,512);
	return (object);
}


void fonts_destroy(struct _font_base *object) {

	if (!object) {
		return;
	}

	free(object->path);
	munmap(object->data,object->size);
	close(object->fd);
	free(object);
}


void fonts_set_color_table(struct _font_base *font, struct ttos_color *color) {

	uint16_t r,g,b,loop,*pointer;

	if ((font->r==color->r)&&(font->g==color->g)&&(font->b==color->b)) {
		return;
	}

	pointer=font->colors;

	for (loop=0;loop<256;loop++) {
		r=(uint16_t)color->r;
		r*=loop;
		g=(uint16_t)color->g;
		g*=loop;
		g>>=5;
		b=(uint16_t)color->b;
		b*=loop;
		b>>=11;
		*(pointer++)=(r&0xF800)|(g&0x07E0)|(b&0x1F);
	}
	font->r=color->r;
	font->g=color->g;
	font->b=color->b;
}

int32_t fonts_get_glyph_data(struct _font_base *font,uint32_t glyph, struct _font_data *font_data) {

	uint8_t *data;
	int8_t *data2;

	if ((glyph>font->last_glyph)||(glyph<font->first_glyph)) {
		font_data->width=0;
		font_data->height=0;
		font_data->addx=0;
		font_data->addy=0;
		font_data->incx=0;
		font_data->incy=0;
		return -1; // glyph out of range
	}

	data=font->glyphs+font->offsets[glyph-font->first_glyph];

	font_data->width=*(data++);
	font_data->height=*(data++);
	data2=(int8_t *)data;
	font_data->addx=*(data2++);
	font_data->addy=*(data2++);
	font_data->incx=*(data2++);
	font_data->incy=*(data2);
	return 0;
}

int32_t fonts_render_glyph(struct _font_base *font,TTOS_Surface *surface,int32_t x, int32_t y,uint32_t glyph,struct _font_data *font_data) {

	uint8_t *alpha;
	uint8_t *data;
	uint16_t *pixels,*color_table;
	uint32_t jump;
	uint32_t loopx,loopy;

	if (!font) {
		return -3;
	}

	if ((glyph>font->last_glyph)||(glyph<font->first_glyph)) {
		return -1; // glyph out of range
	}

	fonts_get_glyph_data(font,glyph,font_data);

	x+=(int32_t)(font_data->addx);
	y-=(int32_t)(font_data->addy);

	if ((x<0)||(y<0)||((x+font_data->width)>(surface->w))||((y+font_data->height)>(surface->h))) {
		return -2; // coordinates out of range
	}

	data=font->glyphs+6+font->offsets[glyph-font->first_glyph];
	pixels=surface->pixels+(x+y*surface->w);
	if (surface->alpha) {
		alpha=surface->alpha+(x+y*surface->w);
	} else {
		alpha=NULL;
	}
	jump=(uint32_t)(surface->w-font_data->width);
	color_table=font->colors;
	for(loopy=0;loopy<font_data->height;loopy++) {
		for(loopx=0;loopx<font_data->width;loopx++) {
			if (!(*data)) {
				data++;
				pixels++;
				if(alpha)
					alpha++;
				continue;
			}
			*(pixels++)=color_table[*data];
			if (alpha)
				*(alpha++)=255-(*data);
			data++;
		}
		pixels+=jump;
		if(alpha)
			alpha+=jump;
	}

	return 0;
}

intptr_t fonts_utf8_to_32(Cstring text,uint32_t *v_glyph) {

	uint32_t size,loop,glyph;
	uint8_t character;

	character=*text;
	if (character<128) {
		*v_glyph=(uint32_t)character;
		return 1;
	} elif (character<192) { // error
		*v_glyph=0;
		return 1;
	} elif (character<224) {
		size=1;
		glyph=(uint32_t)(character&0x1F);
	} elif (character<240) {
		size=2;
		glyph=(uint32_t)(character&0x0F);
	} elif (character<248) {
		size=3;
		glyph=(uint32_t)(character&0x07);
	} elif (character<252) {
		size=4;
		glyph=(uint32_t)(character&0x03);
	} elif (character<254) {
		size=5;
		glyph=(uint32_t)(character&0x01);
	} else {
		*v_glyph=0;
		return 1;
	}
	for(loop=0;loop<size;loop++) {
		text++;
		character=*text;
		if ((character==0)||((character&0xC0)!=0x80)) { // error
			*v_glyph=0;
			return (1+loop);
		}
		glyph=((glyph<<6)&(0xFFFFFFC0))|(character&0x3F);
	}
	*v_glyph=glyph;
	return (1+size);
}

TTOS_Surface *fonts_render_line(struct _font_base *font, Cstring *text, uint32_t flags, uint32_t d_width, struct ttos_color *color) {

	TTOS_Surface *surface;
	int32_t min_x,min_y,max_x,max_y,ominx,ominy,omaxx,omaxy,x,y,start_x,start_y,end_x,end_y,width,height;
	struct _font_data font_data;
	Cstring text_p,text_p2,cut_point;
	bool ended;
	uint32_t glyph;
	bool first,no_space;
	Cstring *text2;

	text2=text;
	min_x=min_y=max_x=x=y=0;
	max_y=**text=='\n' ? font->average_height : 0;
	cut_point=NULL;
	ended=false;
	first=true;
	no_space=false;
	for(text_p=*text;*text_p;) {
		if ((*text_p!=' ')&&(*text_p!='\n')) {
			no_space=true;
		}
		if ((*text_p==' ')&&(no_space)) {
			ominx=min_x;
			ominy=min_y;
			omaxx=max_x;
			omaxy=max_y;
			cut_point=text_p+1;
		}
		if (*text_p=='\n') {
			ominx=min_x;
			ominy=min_y;
			omaxx=max_x;
			omaxy=max_y;
			cut_point=text_p+1;
			ended=true;
			break;
		}
		text_p2=text_p;
		if (flags&TTOS_WIDGET_FONT_LATIN1) {
			glyph=((uint32_t)(*text_p++))&0x000000FF;
		} else {
			text_p+=fonts_utf8_to_32(text_p,&glyph);
		}
		if (fonts_get_glyph_data(font,glyph,&font_data)) {
			continue;
		}
		start_x=x+font_data.addx;
		start_y=y-font_data.addy;
		end_x=start_x+font_data.width-1;
		end_y=start_y+font_data.height-1;
		x+=font_data.incx;
		y+=font_data.incy;
		if (first) {
			min_x=start_x;
			max_x=end_x;
			min_y=start_y;
			if (max_y==0) {
				max_y=end_y;
			}
			first=false;
		} else {
			if (start_x<min_x) {
				min_x=start_x;
			}
			if (start_y<min_y) {
				min_y=start_y;
			}
			if (end_x>max_x) {
				max_x=end_x;
			}
			if (end_y>max_y) {
				max_y=end_y;
			}
		}
		if ((d_width>0)&&((max_x-min_x)>=d_width)) {
			if ((cut_point)&&(flags&TTOS_WIDGET_FONT_WORD_CUT)) {
				min_x=ominx;
				min_y=ominy;
				max_x=omaxx;
				max_y=omaxy;
			}
			if ((!cut_point)||(!(flags&TTOS_WIDGET_FONT_WORD_CUT))) {
				cut_point=text_p2;
			}
			ended=true;
			break;
		}
	}

	if (!ended) {
		cut_point=NULL;
	}

	width=max_x-min_x+1;
	height=max_y-min_y+1;
	surface=blit_create_surface(width,height,true);
	if (!surface) {
		*text=NULL;
		return NULL;
	}
	blit_erase_surface(surface);
	fonts_set_color_table(font,color);
	x=-min_x;
	y=-min_y;

	for(text_p=*text;(*text_p)&&(text_p!=cut_point);) {
		if (flags&TTOS_WIDGET_FONT_LATIN1) {
			glyph=((uint32_t)(*text_p++))&0x000000FF;
		} else {
			text_p+=fonts_utf8_to_32(text_p,&glyph);
		}
		if (fonts_get_glyph_data(font,glyph,&font_data)) {
			continue;
		}
		fonts_render_glyph(font,surface,x,y,glyph,&font_data);
		x+=font_data.incx;
		y+=font_data.incy;
	}
	if (cut_point==*text2) {
		cut_point++;
	}
	*text=cut_point;
	return (surface);
}

TTOS_Surface *fonts_render_label(struct _font_base *font, Cstring text_b, uint32_t flags, uint32_t d_width, struct ttos_color *color) {

	TTOS_Surface *first,*last,*loop,*tmp,*final;
	uint32_t width,height,y,x;
	Cstring text;

	text=text_b;

	if (!(flags&TTOS_WIDGET_FONT_MULTILINE)) {
		return (fonts_render_line(font,&text,flags,d_width,color));
	}

	first=last=tmp=NULL;
	width=height=0;
	do {
		tmp=fonts_render_line(font,&text,flags,d_width,color);
		tmp->private=NULL;
		if (!first) {
			first=tmp;
			last=tmp;
		} else {
			last->private=tmp;
			last=tmp;
		}
		height+=tmp->h+TTOS_FONT_LINE_SEP;
		if (width<tmp->w) {
			width=tmp->w;
		}
	} while(text!=NULL);

	if (height>TTOS_FONT_LINE_SEP) {
		height-=TTOS_FONT_LINE_SEP;
	}

	final=blit_create_surface(width,height,true);
	if (!final) {
		for(loop=first;loop!=NULL;loop=tmp) {
			tmp=(TTOS_Surface *)loop->private; // next in list
			blit_destroy_surface(loop);
		}
		return (NULL);
	}
	blit_erase_surface(final);
	y=0;
	for(loop=first;loop!=NULL;loop=tmp) {
		tmp=(TTOS_Surface *)loop->private; // next in list
		if ((flags&TTOS_WIDGET_FONT_HALIGN_CENTER)&&(width>loop->w)) {
			x=(width-loop->w)/2;
		} else if ((flags&TTOS_WIDGET_FONT_HALIGN_RIGHT)&&(width>loop->w)) {
			x=(width-loop->w);
		} else {
			x=0;
		}
		blit_blend(loop,0,0,loop->w,loop->h,final,x,y,0,0,final->w,final->h,false);
		y+=loop->h+TTOS_FONT_LINE_SEP;
		blit_destroy_surface(loop);
	}
	return(final);
}
