/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * ipc.c
 *
 *  Created on: 27/02/2011
 *      Author: raster
 */

#include <stdlib.h>
#include <string.h>

#include "ipc.h"
#include "other.h"

void ipc_send_subscribers(struct ipc_bus_element *object) {

	if ((object->flags)&TTOS_IPC_OWNER_TO_ALL) {
		minibus_start_message(write_conn,object->port,TTOS_NO_ANSWER);
		minibus_write_long(write_conn,TTOS_SR_UPDATED_BUS_USERS);
		minibus_write_string(write_conn,object->bus_name);
		minibus_write_long(write_conn,object->listeners);
		minibus_server_send(write_conn,object->connection,false);
	}
}

struct ipc_subscriber *ipc_new_subscriber(struct _client_data *client, struct ipc_bus_element *bus) {

	struct ipc_subscriber *object;

	if (!client) {
		return NULL;
	}

	object=malloc(sizeof(struct ipc_subscriber));
	if (!object) {
		return NULL;
	}
	object->bus=bus;
	object->conn=client->conn;
	object->ipc=client->ipc_uid;
	object->bus->listeners++;
	object->bus->modified=true;
	return (object);
}

void ipc_free_subscriber(void *data) {

	struct ipc_subscriber *object=data;

	if (!data) {
		return;
	}
	object->bus->modified=true;
	object->bus->listeners--;
	if (!object->bus->destroying) {
		ipc_send_subscribers(object->bus);
	}
	free(object);

}

struct ipc_bus_element *ipc_new_bus(Cstring bus_name, struct minibus_connection *connection, uint32_t owner, uint32_t port,
									uint32_t flags) {

	struct ipc_bus_element *object;

	object=malloc(sizeof(struct ipc_bus_element));
	if (!object) {
		return NULL;
	}

	object->bus_name=strdup(bus_name);
	object->owner=owner;
	object->flags=flags;
	object->port=port;
	object->connection=connection;
	object->modified=false;
	object->destroying=false;
	object->listeners=0;
	minibus_list_init(&object->subscribers,ipc_free_subscriber,NULL);

	return(object);

}

void ipc_free_bus(void *data) {

	struct ipc_bus_element *object=data;
	struct minibus_list_element *element;
	struct ipc_subscriber *subscriber;

	if (!object) {
		return;
	}

	object->destroying=true;

	minibus_start_message(write_conn,object->port,TTOS_NO_ANSWER);
	minibus_write_long(write_conn,TTOS_SR_BUS_DESTROYED);
	minibus_write_string(write_conn,object->bus_name);
	for(element=object->subscribers.list;element;element=element->next) {
		subscriber=element->data;
		minibus_server_send(write_conn,subscriber->conn,false);
	}

	minibus_list_free(&object->subscribers);
	uid_release_uid(ttos_ipc_uid,object->port);
	free(object->bus_name);
	free(object);

}

/**
 * Registers a bus. Receives a message with the following fields:
 *
 * String: the bus name
 * Uint32: flags (TTOS_IPC_XXX)
 *
 * @param port The port in which the message has been received
 * @param answer_port The port to which an error (or OK) message will be sent. If it's 0, no message will be sent
 * @param connection The client's connection
 * @param user_data Extra parameter used in callbacks
 *
 * @return 0 on success; -1 on error
 */

uint32_t ipc_register_bus(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *user_data) {

	struct minibus_list_element *element;
	struct ipc_bus_element *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	int32_t error;
	Cstring bus_name;
	uint32_t flags,ipc_uid,tmp;

	client=minibus_get_connection_data(connection);

	data=minibus_read_params(connection,"cl",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	bus_name=data[0].string;
	flags=data[1].value32;

	tmp=flags&(TTOS_IPC_OWNER_TO_ALL|TTOS_IPC_ALL_TO_OWNER);
	if ((!tmp)||(tmp==(TTOS_IPC_OWNER_TO_ALL|TTOS_IPC_ALL_TO_OWNER))) { // must have only one of both flags active
		free(data);
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_FLAGS);
		return -1;
	}

	for(element=ipc.list;element;element=element->next) {
		object=element->data;
		if (!strcmp(object->bus_name,bus_name)) {
			free(data);
			other_send_error(connection,answer_port,TTOS_ERROR_BUS_ALREADY_REGISTERED);
			return -1;
		}
	}

	ipc_uid=uid_get_uid(ttos_ipc_uid,TTOS_IPC_UID_MASK);
	if (!ipc_uid) {
		free(data);
		other_send_error(connection,answer_port,TTOS_ERROR_CANT_RESERVE_UID);
		return -1;
	}

	object=ipc_new_bus(bus_name,connection,client->ipc_uid,ipc_uid,flags);
	if (!object) {
		uid_release_uid(ttos_ipc_uid,ipc_uid);
		free(data);
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return -1;
	}

	minibus_list_add(&ipc,object);
	other_send_uid(connection,answer_port,ipc_uid);
	free(data);
	return 0;
}

/**
 * Unregisters a bus. Receives a message with the following fields
 *
 * String: the bus to unregister
 *
 * @param port
 * @param answer_port
 * @param connection
 * @param user_data
 * @return
 */

uint32_t ipc_unregister_bus(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *user_data) {

	struct minibus_list_element *element;
	struct ipc_bus_element *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	int32_t error;
	Cstring bus_name;

	client=minibus_get_connection_data(connection);

	data=minibus_read_params(connection,"c",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	bus_name=data[0].string;

	for(element=ipc.list;element;element=element->next) {
		object=element->data;
		if (!strcmp(object->bus_name,bus_name)) {
			if (object->owner==client->ipc_uid) {
				free(data);
				minibus_list_remove(element,true);
				other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
				return 0;
			} else {
				free(data);
				other_send_error(connection,answer_port,TTOS_ERROR_NO_OWNER);
				return -1;
			}
		}
	}
	free(data);
	other_send_error(connection,answer_port,TTOS_ERROR_INVALID_BUS);
	return -1;
}

/**
 * Subscribes to a bus. Receives a message with the following fields
 *
 * String: the bus to subscribe
 * Int32: the port to which send the answers
 *
 * @param port
 * @param answer_port
 * @param connection
 * @param user_data
 * @return
 */

uint32_t ipc_subscribe_to_bus(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *user_data) {

	struct minibus_list_element *element,*element2;
	struct ipc_bus_element *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	struct ipc_subscriber *subscriber;
	int32_t error;
	Cstring bus_name;

	client=minibus_get_connection_data(connection);

	data=minibus_read_params(connection,"c",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	bus_name=data[0].string;

	for(element=ipc.list;element;element=element->next) {
		object=element->data;
		if (!strcmp(object->bus_name,bus_name)) {
			for(element2=object->subscribers.list;element2;element2=element2->next) {
				subscriber=element2->data;
				if (subscriber->ipc==client->ipc_uid) {
					free(data);
					other_send_error(connection,answer_port,TTOS_ERROR_ALREADY_SUBSCRIBED);
					return -1;
				}
			}
			subscriber=ipc_new_subscriber(client,object);
			minibus_list_add(&object->subscribers,subscriber);
			ipc_send_subscribers(object);
			free(data);
			other_send_uid(connection,answer_port,object->port);
			return 0;
		}
	}
	free(data);
	other_send_error(connection,answer_port,TTOS_ERROR_INVALID_BUS);
	return -1;
}

/**
 * Unsubscribes to a bus. Receives a message with the following fields
 *
 * String: the bus to unsubscribe
 *
 * @param port
 * @param answer_port
 * @param connection
 * @param user_data
 * @return
 */

uint32_t ipc_unsubscribe_to_bus(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *user_data) {

	struct minibus_list_element *element,*element2;
	struct ipc_bus_element *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	struct ipc_subscriber *subscriber;
	int32_t error;
	Cstring bus_name;

	client=minibus_get_connection_data(connection);

	data=minibus_read_params(connection,"c",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	bus_name=data[0].string;

	for(element=ipc.list;element;element=element->next) {
		object=element->data;
		if (!strcmp(object->bus_name,bus_name)) {
			for(element2=object->subscribers.list;element2;element2=element2->next) {
				subscriber=element2->data;
				if (subscriber->ipc==client->ipc_uid) {
					minibus_list_remove(element2,true);
					free(data);
					other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
					return 0;
				}
			}
			free(data);
			other_send_error(connection,answer_port,TTOS_ERROR_NOT_SUBSCRIBED);
			return -1;
		}
	}
	free(data);
	other_send_error(connection,answer_port,TTOS_ERROR_INVALID_BUS);
	return -1;
}
