/*
 * Copyright 2010 (C) Raster Software Vigo (Sergio Costas)
 * This file is part of MINIBUS
 *
 * MINIBUS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * MINIBUS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * queue.h
 *
 *  Created on: 19/08/2010
 *      Author: raster
 */

#include <stdint.h>
#include <stdbool.h>
#include <poll.h>
#include <stddef.h>

#ifndef MINIBUS_H_
#define MINIBUS_H_

typedef char * Cstring;

struct _dstring {

	Cstring text;
	size_t size;
	size_t allocated;

};

typedef struct _dstring * Dstring;

#define elif else if

enum MINIBUS_CONNECTION_STATUS { MINIBUS_CONNECTION_READ, MINIBUS_CONNECTION_WRITE, MINIBUS_CONNECTION_RECEIVING,
								 MINIBUS_CONNECTION_IDLE };
enum MINIBUS_DATA_TYPE {MINIBUS_TYPE_EMPTY, MINIBUS_TYPE_INT8, MINIBUS_TYPE_INT16, MINIBUS_TYPE_INT32,
					  MINIBUS_TYPE_STRING, MINIBUS_TYPE_BLOB, MINIBUS_TYPE_ERROR };

enum MINIBUS_CONNECTION_TYPE {MINIBUS_CLIENT, MINIBUS_LISTENER};

struct minibus_data_element {
	enum MINIBUS_DATA_TYPE type;
	union {
		struct {
			Cstring string;
			uint16_t stringsize;
		};
		struct {
			uint8_t *blob;
			uint16_t blobsize;
		};
		uint16_t value16;
		uint64_t value64;
		uint32_t value32;
		uint8_t value8;
	};
};

struct minibus_list {

	struct minibus_list_element *list;
	struct minibus_list_element *last;
	void (*destroy_callback) (void *element);
	bool (*sort_callback) (void *element1, void *element2);

};

struct minibus_list_element {

	struct minibus_list_element *next;
	struct minibus_list_element *prev;
	void *data;
	struct minibus_list *list;
	uint32_t refcount;
};

struct minibus_connection {

	void (*listener_callback) (int fd, short revents,void *data); ///< listener callback (if this is a listener)
	void *data; ///< user data
	struct pollfd *polling; ///< polling struct used in the loop
	struct minibus_master *master_queue; ///< The queue which manages this connection
	int id_socket; ///< socket for this connection
	enum MINIBUS_CONNECTION_TYPE type;
	enum MINIBUS_CONNECTION_STATUS r_mode;
	enum MINIBUS_CONNECTION_STATUS s_mode;

	struct minibus_list send_buffer;
	struct minibus_list receive_buffer;

	uint32_t s_size; ///< size of the message currently stored
	uint32_t s_pointer; ///< position where the functions are reading or writing parameters
	uint32_t send_remaining; ///< bytes remaining to send from the current message in send_buffer
	uint32_t sent; ///< bytes already sent from the current message in send_buffer
	uint32_t r_pointer,r_size,r_readed,r_total; ///< used during message reception
	short events;
	uint8_t *received_message; ///< pointer to a memory zone where to store a receiving message
	uint8_t isclosing; ///< True if the connection is being closed
	uint8_t *send_message;
};

struct minibus_master {

	int id_socket;
	uint32_t server;
	struct minibus_list conections;
	struct minibus_list callbacks;
	struct minibus_connection *send_connection;
	void *default_data;
	void *new_data;
	void *closed_data;
	void *timeout_data;
	uint32_t (*default_callback)(uint32_t port, uint32_t answer_port, struct minibus_connection *queue, void *data);
	uint32_t (*new_connection_callback)(struct minibus_connection *connection,void *data);
	uint32_t (*closed_connection_callback)(struct minibus_connection *connection,void *data);
	int32_t (*timeout_callback)(struct minibus_master *object,uint32_t timeout,void *data);
	uint32_t timeout_mode;
	struct pollfd *polling;
	int pollsize_total;
	int pollsize_connections;
	bool loop;
	bool forced_exit;
};

struct minibus_wait_data {

	struct minibus_connection *conn;
	uint32_t port;

};

struct minibus_message_element {

	uint32_t size;
	uint32_t event;
	uint32_t count;
	uint32_t pointer;
	uint8_t *data;
};

struct minibus_callback {

	void *data; // user data
	uint32_t port;
	uint32_t (*function)(uint32_t port, uint32_t answer_port, struct minibus_connection *queue, void *data);

};

void minibus_list_init(struct minibus_list *list, void (*destroy_callback) (void *element), bool (*sort_callback) (void *element1, void *element2));
struct minibus_list_element *minibus_list_add(struct minibus_list *list,void *data);
struct minibus_list_element *minibus_list_add_sorted(struct minibus_list *list,void *data);
void minibus_list_remove(struct minibus_list_element *element,bool free_data);
int32_t minibus_list_remove_data(struct minibus_list *list,void *data, bool free_data);
void minibus_list_free(struct minibus_list *list);
uint32_t minibus_list_incref(struct minibus_list_element *object);
uint32_t minibus_list_decref(struct minibus_list_element *object);
int32_t minibus_list_append(struct minibus_list *destination, struct minibus_list *origin);
struct minibus_list_element *minibus_find_data(struct minibus_list *list, void *data);

struct pollfd *minibus_fill_polling(struct minibus_master *object, int serversock);
void minibus_destroy_polling(struct minibus_master *object);

struct minibus_master *minibus_init(char *minibus_path, bool server);
int32_t minibus_release(struct minibus_master *object);

int32_t minibus_accept(struct minibus_master *object);
int32_t minibus_run_loop(struct minibus_master *object);
int32_t minibus_exit_loop(struct minibus_master *object);

void *minibus_get_connection_data(struct minibus_connection *object);
void minibus_set_connection_data(struct minibus_connection *object,void *data);
struct minibus_connection *minibus_get_first_connection(struct minibus_master *object);
int32_t minibus_get_uid(struct minibus_connection *object);

int32_t minibus_receive(struct minibus_connection *object, int32_t msecs);
int32_t minibus_send(struct minibus_connection *object);
int32_t minibus_server_send(struct minibus_connection *from_object,struct minibus_connection *to_object,bool resend);
void minibus_send_reset(struct minibus_connection *object);

int32_t minibus_add_listener(struct minibus_master *object, int fd, short events, void (*listener_callback) (int fd, short revents,void *data), void *data);
int32_t minibus_remove_listener(struct minibus_master *object, int fd);
int32_t minibus_add_callback(struct minibus_master *object, uint32_t port, uint32_t (*function)(uint32_t port, uint32_t answer_port, struct minibus_connection *queue, void *data),void *data);
int32_t minibus_remove_callback(struct minibus_master *object, uint32_t port);
void minibus_set_default_callback(struct minibus_master *object, uint32_t (*function)(uint32_t port, uint32_t answer_port, struct minibus_connection *queue, void *data),void *data);
void minibus_set_new_connection_callback(struct minibus_master *object, uint32_t (*new_connection_callback)(struct minibus_connection *connection,void *data),void *data);
void minibus_set_closed_connection_callback(struct minibus_master *object, uint32_t (*closed_connection_callback)(struct minibus_connection *connection,void *data),void *data);
void minibus_set_timeout_callback(struct minibus_master *object, int32_t (*timeout_callback)(struct minibus_master *object, uint32_t timeout,void *data),void *data);

bool minibus_start_message(struct minibus_connection *connection, uint32_t port, uint32_t answer_port);
int32_t minibus_write_byte(struct minibus_connection *object, uint8_t value);
int32_t minibus_write_short(struct minibus_connection *object, uint16_t value);
int32_t minibus_write_long(struct minibus_connection *object, uint32_t value);
int32_t minibus_write_double(struct minibus_connection *object, uint64_t value);
int32_t minibus_write_string(struct minibus_connection *object, Cstring string);
int32_t minibus_write_blob(struct minibus_connection *object, uint8_t *data, uint32_t size);
int32_t minibus_read_data(struct minibus_connection *object,struct minibus_data_element *data);
struct minibus_data_element *minibus_read_params(struct minibus_connection *object, char *format,int32_t *error);
int32_t minibus_wait_for_port(struct minibus_connection *conn, uint32_t port, bool do_callback);

struct minibus_message_element *minibus_create_send_message_element(struct minibus_connection *object);
struct minibus_message_element *minibus_create_receive_message_element(struct minibus_connection *object, uint32_t event);
void minibus_destroy_message_element (void *object);

int32_t minibus_read_message(struct minibus_master *object,int serversock,struct minibus_wait_data *retval, char wait_for_listeners);
int32_t minibus_do_callback(struct minibus_master *object, struct minibus_wait_data *wdata);

Cstring minibus_path_join(Cstring path1,Cstring path2);

Dstring minibus_new_dstring(char *text);
void minibus_destroy_dstring(Dstring data, bool free_text);
void minibus_append_dstring(Dstring data, char *text);
void minibus_cut_dstring(Dstring data, size_t new_size);

#endif /* MINIBUS_H_ */
